// this will help make a map of cursor movements

#include <iostream>
#include <fstream>
using namespace std;

#define NUMFRAMES	0x0C
#define NUMRERECORD	0x10
#define CTRLDATALEN	0x14
#define SAVESTLOC	0x18
#define CTRLDATALOC	0x1C
#define MD5SUM		0x20	// 16-byte
#define EMUVER		0x30
#define ROMNAME		0x34	// null-terminated
#define NOSEEK		-1

#define CCMASK		0x80
#define DELTAMASK	0x60
#define DATAMASK	0x1F
#define CNUMMASK	0x18
#define BUTTONMASK	0x07

#define SHORTMODE	1
#define RAWMODE		2
#define FRAMEMODE	3
#define LONGFRAMEMODE	4
#define ARTMODE		5
#define SUBMODE		6

#define SECONDS	6010	// we're dealing with centiseconds, to avoid floats
#define	MINUTES	(60*SECONDS)
#define HOURS	(60*MINUTES)
#define MSECOND	((float)SECONDS/1000.0)

static int mode;
static ifstream movie_file;
static char romname[256];
static char cactions[9][16] = {
	"No Action",
	"Reset",
	"Power Cycle",
	"VS Ins Coin",
	"VS DipSwitch 0",
	"FDS Insert",
	"FDS Eject",
	"FDS Select Side",
	"UNKNOWN ACTION"};
static char buttons[] = "ABsS^v<>";
static int pbuttons = 0;
static int framecount = 0;
static int statecount = 0;
static int curx,cury,curc;
static int minx,maxx,miny,maxy;
static int pass2 = 0;

static char imgname[256];
static int xsize,ysize;

void print_usage()
{
  printf("Usage: maprom [mode] filename\n");
  printf("  Modes:\n");
  printf("    -summary\tShow header information from the movie file.  Default.\n");
  printf("    -frames\tShow controller state for frames where button state changes\n");
  printf("    -longframes\tShow controller state for EVERY frame\n");
  printf("    -rawdata\tShow result for each byte of controller data\n");
  printf("    -artmode\tPrints script to execute a series of ImageMagick operations\n");
  printf("    -srtsub\tGenerates subtitles in SRT format\n");
}

void safe_printchar(char c)
{
  if (!(c>=33 && c<=126))
    c = '.';
  printf("%c",c);
}

char* echo()
{
  if (mode == ARTMODE)
    return "echo ";
  else
    return "";
}

char* get_caction(int c)
{
  switch (c & DATAMASK)
  {
    case 0:
      return cactions[0];
    case 1:
      return cactions[1];
    case 2:
      return cactions[2];
    case 7:
      return cactions[3];
    case 8:
      return cactions[4];
    case 24:
      return cactions[5];
    case 25:
      return cactions[6];
    case 26:
      return cactions[7];
    default:
      return cactions[8];
  }
}

void get_romname()
{
  int x,c;
  movie_file.seekg(ios::beg);
  movie_file.ignore(ROMNAME);
  for (x=0; (c!='\0') && (x<(int)sizeof(romname)/4) ;x++)
  {
    c = movie_file.get();
    romname[x] = c;
    if ((c >= 48 && c<=57) || (c >= 65 && c <= 90) || (c >= 97 && c <= 122))
      imgname[x] = c;
    else
      imgname[x] = '_';
  }
  if (x < 5)	// if the rom name is too short, it's probably nothing at all
  {
    strcpy(imgname,"route.png");
  }
  else	// otherwise, add an extension.  I should learn more about strings.
  {
    imgname[x-1] = '.';
    imgname[x] = 'p';
    imgname[x+1] = 'n';
    imgname[x+2] = 'g';
    imgname[x+3] = '\0';
  }
}

int get_bytes(int startpos, int len)
// use negative startpos to continue from current position
{
  int x,c,ret;
  if (startpos >= 0)
  {
    movie_file.seekg(ios::beg);
    movie_file.ignore(startpos);
  }
  for (x=0,ret=0;x<len;x++)
  {
    c = movie_file.get();
    ret = ret | c<<(8*x);
  }
  return ret;
}

int get_delta_len(int c)
{
  return ((c & DELTAMASK) >> 5);
}

void toggle_button(int c)
{
  //printf("T: %2x (%2x) ",pbuttons,(1 <<(c&BUTTONMASK)));
  pbuttons = pbuttons ^ (1 << (c&BUTTONMASK));
  //printf("%2x\n",pbuttons);
  if (mode == ARTMODE)
  {
    switch(c&BUTTONMASK)
    {
      case 0:
        curc = curc ^ 0xf00;
	break;
      case 1:
        curc = curc ^ 0x00f;
	break;
      case 2:
        curc = curc ^ 0x404;
	break;
      case 3:
        curc = curc ^ 0x0c0;
	break;
    }
  }
}

void print_full_buttons(int c, int d)
{
  int x,y;
  //printf("%2x ",pbuttons);
  for (y=0; y<d; y++)
  {
    framecount++;
    printf("%8d ",framecount);
    for (x=0; x<8; x++)
    {
      if (pbuttons & (1<<(x&BUTTONMASK)))
        printf("%c",buttons[x]);
      else
        printf("-");
    }
    printf("\n");
  }
  if (c>=0)
    toggle_button(c);
}

void print_short_buttons(int c, int d)
{
  int x,y;
  //printf("%2x ",pbuttons);
  for (y=0; y<d; y++)
  {
    framecount++;
    if (!y)
    {
      printf("%8d ",framecount);
      for (x=0; x<8; x++)
      {
        if (pbuttons & (1<<(x&BUTTONMASK)))
          printf("%c",buttons[x]);
        else
          printf("-");
      }
    }
    if (!(y%8))
      printf("\n\t ");
    printf(".");
  }
  if (d)
    printf("\n");
  if (c>=0)
    toggle_button(c);
}

void print_coords(int c, int d)
{
  int y,sy,sx;
  sx = curx;
  sy = cury;
  for (y=0; y<d; y++)
  {
    framecount++;
    if(pbuttons&(1<<4))
      cury--;
    if(pbuttons&(1<<5))
      cury++;
    if(pbuttons&(1<<6))
      curx--;
    if(pbuttons&(1<<7))
      curx++;
    if (pass2 && y==d-1)
    {
      printf("echo \'Processing frame %d\'\n",framecount);
      printf("convert %s -strokewidth %d -draw \"stroke #%03x path \'M %d,%d L %d,%d\'\" %s\n",
		imgname,(curc?2:1),curc,sx,sy,curx,cury,imgname);
    }
    if (curx < minx)
      minx = curx;
    if (curx > maxx)
      maxx = curx;
    if (cury < miny)
      miny = cury;
    if (cury > maxy)
      maxy = cury;
  }
  if (c>=0)
    toggle_button(c);
}

void print_subs(int c, int d)
{
  int x,sy,sx;
  int h1,h2,m1,m2,s1,s2,ms1,ms2;
  int startfc, endfc;
  sx = curx;
  sy = cury;
  startfc = framecount*100;
  framecount += d;
  endfc = framecount*100;
  if (d && pbuttons)
  {
/*
    // is there a way to use a%b with non-integers?
    h1 = (int)(startfc/HOURS);
    m1 = (int)((startfc-(int)((int)(h1*HOURS)))/MINUTES);
    s1 = (int)((startfc-(int)((int)(h1*HOURS)+(int)(m1*MINUTES)))/SECONDS);
    ms1 = (int)((startfc-(int)((int)(h1*HOURS)+(int)(m1*MINUTES)+(int)(s1*SECONDS)))*FIMS);
    h2 = (int)(endfc/HOURS);
    m2 = (int)((endfc-(int)((int)(h2*HOURS)))/MINUTES);
    s2 = (int)((endfc-(int)((int)(h2*HOURS)+(int)(m2*MINUTES)))/SECONDS);
    ms2 = (int)((endfc-(int)((int)(h2*HOURS)+(int)(m2*MINUTES)+(int)(s2*SECONDS)))*FIMS);
*/
    h1 = startfc/HOURS;
    m1 = (startfc%HOURS)/MINUTES;
    s1 = (startfc%MINUTES)/SECONDS;
    ms1 = (int)((startfc%SECONDS)/MSECOND);
    h2 = endfc/HOURS;
    m2 = (endfc%HOURS)/MINUTES;
    s2 = (endfc%MINUTES)/SECONDS;
    ms2 = (int)((endfc%SECONDS)/MSECOND);
    printf("%d\n%02d:%02d:%02d,%03d --> %02d:%02d:%02d,%03d\n",
	    ++statecount,h1,m1,s1,ms1,h1,m2,s2,ms2);
    //printf("%d %d %d\n",(int)startfc,(int)endfc,framecount);
    //printf("%d\n",(int)((endfc-(int)((int)(h1*HOURS)+(int)(m1*MINUTES)+(int)(s1*SECONDS)))*FPMS));
    for (x=0; x<8; x++)
    {
      if (pbuttons & (1<<(x&BUTTONMASK)))
        printf("%c  ",buttons[x]);
    }
    printf("\n\n");
  }
  if (c>=0)
    toggle_button(c);
}

void print_buttons(int c,int delta)
{
  switch(mode)
  {
    case LONGFRAMEMODE:
      print_full_buttons(c,delta);
      break;
    case FRAMEMODE:
      print_short_buttons(c,delta);
      break;
    case ARTMODE:
      print_coords(c,delta);
      break;
    case SUBMODE:
      print_subs(c,delta);
      break;
  }
  if (c<0 && mode!=ARTMODE)
    printf("%8d %s\n",framecount,get_caction(-c));
}

void translate_cdata(int startpos, int len)
{
  int x,c,p,dn,delta;
  int ndel = 0;
  delta = dn = 0;
  if (startpos >= 0)
  {
    movie_file.seekg(ios::beg);
    movie_file.ignore(startpos);
  }
  for (x=0;x<len;x++)
  {
    p = movie_file.tellg();
    c = movie_file.get();
    printf("%08X ",p);
    if (ndel)
    {
      delta = delta | c<<(8*dn++);
      printf(" %02X\t",c);
      printf(" Wait");
      if (!--ndel)
      {
	printf(" %d frames first!",delta);
	delta = 0;
	dn = 0;
      }
      else
      {
	printf("..");
      }
      printf("\n");
    }
    else
    {
      ndel = get_delta_len(c);
      printf("%02X\t",c);
      printf("%d Delta bytes, ",ndel);
      if (c & CCMASK)
      {
	printf("System %s",get_caction(c));
      }
      else
      {
	printf("Controller %d, ",((c&CNUMMASK)>>3)+1);
	printf("Toggle %c",buttons[c&BUTTONMASK]);
      }
      printf("\n");
    }
  }
}

void read_cdata(int startpos, int len)
{
  int x,c,delta;
  int ndel = 0;
  if (startpos >= 0)
  {
    movie_file.seekg(ios::beg);
    movie_file.ignore(startpos);
  }
  if (mode==ARTMODE && pass2)
  {
    printf("echo \'Initializing image: %dx%d\'\n",xsize,ysize);
    printf("convert -size %dx%d xc:white %s\n",xsize,ysize,imgname);
  }
  for (x=0;x<len;x++)
  {
    delta = 0;
    ndel = 0;
    c = movie_file.get();
    ndel = get_delta_len(c);
    if (ndel)
    {
      delta = get_bytes(NOSEEK, ndel);
      x += ndel;
    }
    if (c & CCMASK)
    {
      //printf("%2x %4d\n%s\n",c,delta,cactions[c&DATAMASK]);
      if (mode != SUBMODE)
        print_buttons(-c,delta);
      continue;
    }
    // only take player 1 input
    if (c & CNUMMASK)
      continue;
    //printf("%2x %4d\n",c,delta);
    print_buttons(c,delta);
  }
}

int main(int argc, char* argv[])
{
  int ix,iy;
  int numframes,rerecords,cdatalen,savestloc,cdataloc,emuver;

  curc = 0;
  curx = minx = maxx = 0;
  cury = miny = maxy = 0;
  xsize = 0;
  ysize = 0;

  if (argc < 2)
  {
    print_usage();
    return 1;
  }

  if ((argc>=3) && !strcmp(argv[1],"-frames"))
    mode = FRAMEMODE;
  else if((argc>=3) && !strcmp(argv[1],"-longframes"))
    mode = LONGFRAMEMODE;
  else if((argc>=3) && !strcmp(argv[1],"-artmode"))
    mode = ARTMODE;
  else if((argc>=3) && !strcmp(argv[1],"-rawdata"))
    mode = RAWMODE;
  else if((argc>=3) && !strcmp(argv[1],"-srtsub"))
    mode = SUBMODE;
  else if((argc<3) || !strcmp(argv[1],"-summary"))
    mode = SHORTMODE;
  else
  {
    print_usage();
    return 1;
  }

  //movie_file.open(movie_filename, ios::binary);
  movie_file.open(argv[argc-1], ios::binary);
  if (!movie_file)
  {
    print_usage();
    return 1;
  }

  numframes = get_bytes(NUMFRAMES, 4);
  rerecords = get_bytes(NOSEEK, 4);
  cdatalen = get_bytes(NOSEEK, 4);
  savestloc = get_bytes(NOSEEK, 4);
  cdataloc = get_bytes(NOSEEK, 4);
  emuver = get_bytes(EMUVER, 4);

  get_romname();

  if (mode != SUBMODE)
  {
    printf("%s\'File: %s\'\n%s\'Rom: %s\'\n",echo(),argv[1],echo(),romname);
    printf("%s\'fr: %d rr: %d cle: %X slo: %X clo: %X v: %d\'\n",
	echo(),numframes, rerecords, cdatalen, savestloc, cdataloc, emuver);
  }
  else
  {
    framecount = 61;	// to offset the TAS intro
  }

  if (mode>RAWMODE)
    read_cdata(cdataloc, cdatalen);
  else if (mode == RAWMODE)
    translate_cdata(cdataloc, cdatalen);

  if (mode==ARTMODE)
  {
    pass2 = 1;
    framecount = 0;
    statecount = 0;
    pbuttons = 0;
    xsize = -minx + maxx + 1;
    ysize = -miny + maxy + 1;
    curx = minx = maxx = ix = -minx;
    cury = miny = maxy = iy = -miny;
    curc = 0;
    read_cdata(cdataloc, cdatalen);
    printf("echo 'Adding start/end indicators'\n");
    printf("convert %s -resize %dx%d\\< -size %dx%d xc:white +swap -gravity north-east -composite -draw \"path \'M 7,0 L 7,%d\'\" -draw \"path \'M 7,%d L %d,%d\'\" -draw \"path \'M 7,%d L 0,%d\'\" -draw \"path \'M %d,%d L %d,%d\'\" -draw \"path \'M 7,%d L 0,%d\'\" -draw \"path \'M %d,%d L %d,%d\'\" %s\n",
		imgname,xsize,ysize,xsize+10,ysize+10,
		ysize+3,ysize+3,xsize+10,ysize+3,
		iy,iy,ix+10,ysize+3,ix+10,ysize+10,
		cury,cury,curx+10,ysize+3,curx+10,ysize+10,imgname);
    printf("convert %s -resize %dx%d\\< -size %dx%d xc:white +swap -gravity center -composite %s\n",
		imgname,xsize,ysize,xsize+20,ysize+20,imgname);
  }
 
  movie_file.close();
  return 0;
}

